import logging
from collections import defaultdict
from pathlib import Path
from types import SimpleNamespace
from typing import Any, Tuple

import numpy as np
import pandas as pd
from skimage.measure import label, regionprops
from spotipy.utils import points_matching
from stardist.models import StarDist2D

from cenfind.core.data import Dataset, Field
from cenfind.core.detectors import extract_foci, extract_nuclei
from cenfind.core.outline import Centre, draw_foci, Contour

import cv2


def signed_distance(focus: Centre, nucleus: Contour) -> float:
    """Wrapper for the opencv PolygonTest"""

    result = cv2.pointPolygonTest(nucleus.contour, focus.to_cv2(), measureDist=True)
    return result


def full_in_field(coordinate, image_shape, fraction) -> bool:
    h, w = image_shape
    pad_lower = int(fraction * h)
    pad_upper = h - pad_lower
    if all([pad_lower < c < pad_upper for c in coordinate]):
        return True
    return False


def flag(is_full: bool) -> tuple:
    return (0, 255, 0) if is_full else (0, 0, 255)


def infer_centrosomes(foci: list, img_shape: tuple, distance=0.6) -> list:
    """
    Add centrosome to a list of foci.
    Args:
        foci:
        img_shape:
        distance: twice the inter-centriole distance in micrometres (2 x .3 um)

    Returns:

    """
    _foci = foci.copy()

    centrosomes_mask = np.zeros(img_shape, dtype="uint8")
    centrosomes_mask = draw_foci(centrosomes_mask, _foci, radius=distance)

    centrosomes_map = label(centrosomes_mask)
    centrosomes_centroids = regionprops(centrosomes_map)

    for f in _foci:
        foci_index = centrosomes_map[f.centre]
        centrosome_centroid = centrosomes_centroids[foci_index - 1].centroid
        centrosome_centroid = tuple(int(c) for c in centrosome_centroid)
        f.parent = Centre(centrosome_centroid, label="Centrosome")

    return foci


def assign(
    foci: list, nuclei: list, vicinity: float, pixel_size: float
) -> list[tuple[Any, list[Any]]]:
    """
    Assign centrioles to nuclei in one field
    :param foci
    :param nuclei
    :param vicinity: the distance in pixels, below which centrioles are assigned
     to nucleus
    :param pixel_size: in micrometres
    :return: List[Tuple[Centre, Contour]]

    """
    if len(nuclei) == 0:
        raise ValueError("foci is an empty list")
    _foci = foci.copy()
    _nuclei = nuclei.copy()

    vicinity_pixel = int(vicinity / pixel_size)

    nuclei_pos = [tuple(n.centre.to_numpy()) for n in _nuclei]

    # Initialise the pairs so that nuclei
    # with no centrioles are maintained in the output
    container = defaultdict(list)
    for k in nuclei_pos:
        container.setdefault(k, [])

    # Take each focus and compute the distance to all nuclei
    # and assign it to the nearest nucleus, if it is in the vicinity
    while len(_foci):
        f = _foci.pop()
        centrosome = f.parent
        dists = []
        for n in _nuclei:
            nuclei_pos = tuple(n.centre.to_numpy())
            dist = signed_distance(centrosome, n)
            dists.append((nuclei_pos, dist))
        nuclei_pos_nearest = max(dists, key=lambda t: t[1])
        nearest_nucleus, max_signed_dist = nuclei_pos_nearest
        if max_signed_dist > vicinity_pixel:
            container[nearest_nucleus].append(f)
            continue

    pairs = [(k, v) for k, v in container.items()]

    return pairs


# TODO: refactor as a field method
def field_score(
    field: Field,
    model_nuclei: StarDist2D,
    model_foci: Path,
    nuclei_channel: int,
    factor,
    vicinity,
    channel: int,
) -> Tuple[np.ndarray, list]:
    """
    1. Detect foci in the given channels
    2. Detect nuclei
    3. Assign foci to nuclei
    :param channel:
    :param nuclei_channel:
    :param model_foci:
    :param model_nuclei:
    :param field:
    :return: list(foci, nuclei, assigned, scores)
    """

    image_shape = field.projection.shape[1:]
    centres, nuclei = extract_nuclei(field, nuclei_channel, factor, model_nuclei)
    if len(nuclei) == 0:
        raise ValueError("No nucleus has been detected")
    prob_map, foci = extract_foci(
        data=field, foci_model_file=model_foci, channel=channel
    )
    foci = [Centre((r, c), f_id, "Centriole") for f_id, (r, c) in enumerate(foci)]

    foci = infer_centrosomes(foci, image_shape, distance=0.6)
    assigned = assign(foci=foci, nuclei=nuclei, vicinity=vicinity, pixel_size=0.1025)

    scores = []
    for pair in assigned:
        nucleus, focus = pair
        scores.append(
            {
                "fov": field.name,
                "channel": channel,
                "nucleus": nucleus,
                "score": len(focus),
                "is_full": full_in_field(nucleus, image_shape, 0.05),
            }
        )
    return foci, nuclei, assigned, scores


# TODO: refactor as a fiield method
def field_metrics(
    field: Field,
    channel: int,
    annotation: np.ndarray,
    predictions: np.ndarray,
    tolerance: int,
    threshold: float,
) -> dict:
    """
    Compute the accuracy of the prediction on one field.
    :param threshold:
    :param field:
    :param channel:
    :param annotation:
    :param predictions:
    :param tolerance:
    :return: dictionary of fields
    """
    if all((len(predictions), len(annotation))) > 0:
        res = points_matching(annotation, predictions, cutoff_distance=tolerance)
    else:
        logging.warning(
            "threshold: %f; detected: %d; annotated: %d... Set precision and accuracy to zero"
            % (threshold, len(predictions), len(annotation))
        )
        res = SimpleNamespace()
        res.precision = 0.0
        res.recall = 0.0
        res.f1 = 0.0
        res.tp = (0,)
        res.fp = (0,)
        res.fn = 0
    perf = {
        "dataset": field.dataset.path.name,
        "field": field.name,
        "channel": channel,
        "n_actual": len(annotation),
        "n_preds": len(predictions),
        "threshold": threshold,
        "tolerance": tolerance,
        "tp": res.tp[0] if type(res.tp) == tuple else res.tp,
        "fp": res.fp[0] if type(res.fp) == tuple else res.fp,
        "fn": res.fn[0] if type(res.fn) == tuple else res.fn,
        "precision": np.round(res.precision, 3),
        "recall": np.round(res.recall, 3),
        "f1": np.round(res.f1, 3),
    }
    return perf


# TODO: refactor as a field method
def dataset_metrics(
    dataset: Dataset, split: str, model: Path, tolerance, threshold
) -> tuple[dict, list]:
    perfs = []
    prob_maps = {}
    for field, channel in dataset.pairs(split):
        annotation = field.annotation(channel)
        prob_map, predictions = extract_foci(
            field, model, channel, prob_threshold=threshold
        )
        perf = field_metrics(
            field, channel, annotation, predictions, tolerance, threshold=threshold
        )
        perfs.append(perf)
        prob_maps[field.name] = prob_map
    return prob_maps, perfs


def field_score_frequency(df, by="field"):
    """
    Count the absolute frequency of number of centriole per well or per field
    :param df: Df containing the number of centriole per nuclei
    :return: Df with absolut frequencies.
    """
    cuts = [0, 1, 2, 3, 4, 5, np.inf]
    labels = "0 1 2 3 4 +".split(" ")

    df = df.set_index(["fov", "channel"])
    result = pd.cut(df["score"], cuts, right=False, labels=labels, include_lowest=True)
    result = result.groupby(["fov", "channel"]).value_counts()
    result.name = "freq_abs"
    result = result.sort_index().reset_index()
    result = result.rename({"score": "score_cat"}, axis=1)
    if by == "well":
        result[["well", "field"]] = result["fov"].str.split("_", expand=True)
        print(result.columns)
        result = result.groupby(["well", "channel", "score_cat"])[["freq_abs"]].sum()
        result = result.reset_index()
        result = result.pivot(index=["well", "channel"], columns="score_cat")
        result.reset_index().sort_values(["channel", "well"])
    else:
        result = result.groupby(["fov", "channel", "score_cat"]).sum()
        result = result.reset_index()
        result = result.pivot(index=["fov", "channel"], columns="score_cat")
        result.reset_index().sort_values(["channel", "fov"])

    return result


def save_foci(foci_list: list[Centre], dst: str, logger=None) -> None:
    if len(foci_list) == 0:
        array = np.array([])
        if logger is not None:
            logger.info("No centriole detected")
        else:
            print("No centriole detected")
    else:
        array = np.asarray(np.stack([c.to_numpy() for c in foci_list]))
        array = array[:, [1, 0]]
    np.savetxt(dst, array, delimiter=",", fmt="%u")



import logging
from datetime import datetime
from pathlib import Path

# TODO: Implement logger objects
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
start_stamp = datetime.now()
log_file = f'{start_stamp.strftime("%Y%m%d_%H:%M:%S")}_train.log'
fh = logging.FileHandler(Path("./logs") / log_file)
fh.setFormatter(formatter)
logger.addHandler(fh)



import sys
from pathlib import Path

import pandas as pd

from cenfind.core.data import Dataset


def register_parser(parent_subparsers):
    parser = parent_subparsers.add_parser(
        "evaluate", help="Evaluate the model on the test split of the dataset"
    )

    parser.add_argument("dataset", type=Path, help="Path to the dataset folder")
    parser.add_argument("model", type=Path, help="Path to the model")
    parser.add_argument(
        "--performances_file",
        type=Path,
        help="Path of the performance file, STDOUT if not specified",
    )
    parser.add_argument(
        "--tolerance",
        type=int,
        default=3,
        help="Distance in pixels below which two points are deemed matching",
    )

    return parser


def run(args):
    dataset = Dataset(args.dataset)
    if not any(dataset.path_annotations_centrioles.iterdir()):
        print(f"ERROR: The dataset {dataset.path.name} has no annotation. You can run `cenfind predict` instead")
        sys.exit(2)

    from cenfind.core.measure import dataset_metrics
    _, performance = dataset_metrics(
        dataset, split="test", model=args.model, tolerance=args.tolerance, threshold=.5
    )

    performance_df = pd.DataFrame(performance)
    performance_df = performance_df.set_index("field")
    if args.performances_file:
        performance_df.to_csv(args.performance_file)
    else:
        print(performance_df)



import contextlib
import os
from pathlib import Path
from typing import Tuple, List
import functools

import cv2
import numpy as np
import tensorflow as tf
from csbdeep.utils import normalize
from spotipy.utils import normalize_fast2d
from spotipy.model import SpotNet
from stardist.models import StarDist2D

from cenfind.core.data import Field
from cenfind.core.outline import resize_image
from cenfind.core.outline import Centre, Contour

np.random.seed(1)
tf.random.set_seed(2)


@functools.lru_cache(maxsize=None)
def get_model(model):
    path = Path(model)
    if not path.is_dir():
        raise (FileNotFoundError(f"{path} is not a directory"))

    return SpotNet(None, name=path.name, basedir=str(path.parent))


def extract_foci(
    data: Field,
    foci_model_file: Path,
    channel: int,
    prob_threshold=0.5,
    min_distance=2,
    **kwargs,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Detect centrioles as row, col, row major
    :param data:
    :param foci_model_file:
    :param channel:
    :param prob_threshold:
    :param min_distance:
    :param kwargs:
    :return:
    """
    data = data.channel(channel)
    with open(os.devnull, "w") as f, contextlib.redirect_stdout(f):
        data = normalize_fast2d(data)
        model = get_model(foci_model_file)
        mask_preds, points_preds = model.predict(
            data, prob_thresh=prob_threshold, min_distance=min_distance, verbose=False
        )
    return mask_preds, points_preds


def extract_nuclei(
    field: Field, channel: int, factor: int, model: StarDist2D = None, annotation=None
) -> Tuple[List[Centre], List[Contour]]:
    """
    Extract the nuclei from the nuclei image
    :param field:
    :param channel:
    :param factor: the factor related to pixel size
    :param model:
    :param annotation: a mask with pixels labelled for each centre

    :return: List of Contours.

    """

    if annotation is not None:
        nuclei_detected = annotation
    elif model is not None:
        data = field.channel(channel)
        data_resized = resize_image(data, factor)
        with open(os.devnull, "w") as f, contextlib.redirect_stdout(f):
            labels, coords = model.predict_instances(normalize(data_resized))
        nuclei_detected = cv2.resize(
            labels, dsize=data.shape, fx=1, fy=1, interpolation=cv2.INTER_NEAREST
        )

    else:
        raise ValueError("Please provide either an annotation or a model")

    labels_id = np.unique(nuclei_detected)

    cnts = []
    for nucleus_id in labels_id:
        if nucleus_id == 0:
            continue
        sub_mask = np.zeros_like(nuclei_detected, dtype="uint8")
        sub_mask[nuclei_detected == nucleus_id] = 255
        contour, hierarchy = cv2.findContours(
            sub_mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
        )
        cnts.append(contour[0])

    contours = tuple(cnts)
    contours = [
        Contour(c, "Nucleus", c_id, confidence=-1) for c_id, c in enumerate(contours)
    ]

    centres = [c.centre for c in contours]

    return centres, contours



import sys
from pathlib import Path
from cenfind.core.data import Dataset


def register_parser(parent_subparsers):
    parser = parent_subparsers.add_parser("squash", help="Write z-projections.")
    parser.add_argument("path", type=Path, help="Path to the dataset folder")

    return parser


def run(args):
    """
    Squash the raw fields of view.
    """

    dataset = Dataset(args.path)

    if dataset.has_projections:
        print("Projections already exist, squashing skipped.")
        sys.exit()

    if not dataset.raw.exists():
        print(
            "ERROR: Folder raw/ does not exist. Make sure to move all your images under raw/"
        )
        sys.exit()

    dataset.write_projections()
    print("Projections saved under %s" % dataset.projections)



import os
import sys
import contextlib
from pathlib import Path

from tqdm import tqdm
import pandas as pd
import tifffile as tif

from cenfind.core.data import Dataset
from cenfind.core.measure import save_foci
from cenfind.core.outline import save_visualisation


def register_parser(parent_subparsers):
    parser = parent_subparsers.add_parser(
        "score", help="Score the projections given the channels"
    )
    parser.add_argument("dataset", type=Path, help="Path to the dataset")
    parser.add_argument("model", type=Path, help="Absolute path to the model folder")
    parser.add_argument(
        "--channel_nuclei",
        type=int,
        required=True,
        help="Channel index for nuclei segmentation, e.g., 0 or 3",
    )
    parser.add_argument(
        "--channel_centrioles",
        nargs="+",
        type=int,
        required=True,
        help="Channel indices to analyse, e.g., 1 2 3",
    )
    parser.add_argument(
        "--vicinity",
        type=int,
        default=-5,
        help="Distance threshold in micrometer (default: -5 um)",
    )
    parser.add_argument(
        "--factor",
        type=int,
        default=256,
        help="Factor to use: given a 2048x2048 image, 256 if 63x; 2048 if 20x:",
    )
    parser.add_argument("--cpu", action="store_true", help="Only use the cpu")

    return parser


def run(args):
    if not Path(args.model).exists():
        raise FileNotFoundError(f"{args.model} does not exist")

    if args.channel_nuclei in set(args.channel_centrioles):
        raise ValueError("Nuclei channel cannot be in channels")

    if args.cpu:
        os.environ["CUDA_VISIBLE_DEVICES"] = "-1"

    dataset = Dataset(args.dataset)

    if not any(dataset.projections.iterdir()):
        print(
            "The projection folder (%s) is empty.\nPlease ensure you have run `squash` or that you have put the projections under projections/"
            % dataset.projections
        )
        sys.exit()

    channels = dataset.fields[0].projection.shape[0]

    if args.channel_nuclei not in range(channels):
        print("Index for nuclei (%s) out of index range" % args.channel_nuclei)
        sys.exit()

    if not set(args.channel_centrioles).issubset(set(range(channels))):
        print(
            "Channels (%s) out of channel range %s" % args.channel_centrioles,
            set(range(channels)),
        )
        sys.exit()

    from stardist.models import StarDist2D

    with open(os.devnull, "w") as f, contextlib.redirect_stdout(f):
        model_stardist = StarDist2D.from_pretrained("2D_versatile_fluo")

    scores = []

    pbar = tqdm(dataset.fields)
    from cenfind.core.measure import field_score

    for field in pbar:
        print("Processing %s" % field.name)
        pbar.set_description(f"{field.name}")

        for ch in args.channel_centrioles:
            print("Processing %s / %d" % (field.name, ch))
            try:
                foci, nuclei, assigned, score = field_score(
                    field=field,
                    model_nuclei=model_stardist,
                    model_foci=args.model,
                    nuclei_channel=args.channel_nuclei,
                    factor=args.factor,
                    vicinity=args.vicinity,
                    channel=ch,
                )
                predictions_path = (
                    dataset.predictions
                    / "centrioles"
                    / f"{field.name}{dataset.projection_suffix}_C{ch}.txt"
                )
                save_foci(foci, predictions_path)
                print(
                    "(%s), channel %s: nuclei: %s; foci: %s"
                    % (field.name, ch, len(nuclei), len(foci))
                )
                pbar.set_postfix(
                    {
                        "field": field.name,
                        "channel": ch,
                        "nuclei": len(nuclei),
                        "foci": len(foci),
                    }
                )
                scores.append(score)

                print("Writing visualisations for (%s), channel %s" % (field.name, ch))
                vis = save_visualisation(
                    field, foci, ch, nuclei, args.channel_nuclei, assigned
                )
                tif.imwrite(dataset.visualisation / f"{field.name}_C{ch}_pred.png", vis)

            except ValueError as e:
                print("%s (%s)" % (e, field.name))
                continue

        print("DONE (%s)" % field.name)

    flattened = [leaf for tree in scores for leaf in tree]
    scores_df = pd.DataFrame(flattened)
    scores_df.to_csv(dataset.statistics / f"scores_df.tsv", sep="\t", index=False)
    print("Writing raw scores to %s" % dataset.statistics / f"scores_df.tsv")



import sys
from pathlib import Path
from cenfind.core.data import Dataset
import pytomlpp


def register_parser(parent_subparsers):
    parser = parent_subparsers.add_parser(
        "prepare", help="Prepare the dataset structure"
    )
    parser.add_argument("dataset", type=Path, help="Path to the dataset")
    parser.add_argument(
        "--projection_suffix",
        type=str,
        default="",
        help="Suffix indicating projection, e.g., `_max` or `Projected`, empty if not specified",
    )
    parser.add_argument(
        "--splits",
        type=int,
        nargs="+",
        help="Write the train and test splits for continuous learning using the channels specified",
    )

    return parser


def run(args):
    """Set up the dataset folder with default directories"""

    if not args.dataset.exists():
        print(f"The path `{args.path}` does not exist.")
        sys.exit()

    with open(args.dataset / "metadata.toml", "w") as f:
        content = pytomlpp.dumps({"projection_suffix": args.projection_suffix})
        f.write(content)

    dataset = Dataset(args.dataset, projection_suffix=args.projection_suffix)

    dataset.setup()
    dataset.write_fields()
    if args.splits:
        dataset.write_train_test(args.splits)



import pandas as pd
from matplotlib import pyplot as plt

from cenfind.core.constants import ROOT_DIR
from cenfind.core.constants import protein_positions
pd.options.display.max_columns =50

def main():
    path_data = ROOT_DIR / 'out/performances_20221006_130126.csv'
    data = pd.read_csv(path_data)
    data = data.loc[data['n_preds'] > 0]

    datasets_unique = data.dataset.unique()

    fig, axs = plt.subplots(1, len(datasets_unique), figsize=(5*len(datasets_unique), 5))
    for r, dataset in enumerate(datasets_unique):
        sub = data.loc[data['dataset'] == dataset]
        ax = axs[r]
        for channel in sub.channel.unique():
            marker = protein_positions[dataset][channel]

            ssub = sub.loc[sub['channel'] == channel]
            ssub = ssub.sort_values(by='threshold')
            ax.plot(ssub['recall'], ssub['precision'], ls='-', label=marker)

        ax.spines.right.set_visible(False)
        ax.spines.top.set_visible(False)
        ax.set_xlim(0, 1)
        ax.set_ylim(0, 1.1)
        ax.set_xlabel('recall')
        ax.set_ylabel('precision')
        ax.set_title(dataset)
        ax.legend()
    fig.savefig(ROOT_DIR / f'out/checks/precision_recall.png')

if __name__ == '__main__':
    main()
